// Abstract Syntax Tree for Futil
use crate::errors::{Error, Result, Span};
use crate::lang::{context::LibraryContext, structure::StructureGraph};
use derivative::Derivative;
use itertools::Itertools;
use petgraph::graph::NodeIndex;
use std::collections::HashMap;
use std::hash::Hash;
use std::ops::{BitAnd, BitOr, Not};

/// Represents an identifier in a Futil program
#[derive(Derivative, Clone, PartialOrd, Ord)]
#[derivative(Hash, Eq, PartialEq)]
pub struct Id {
    pub id: String,
    #[derivative(Hash = "ignore")]
    #[derivative(PartialEq = "ignore")]
    span: Option<Span>,
}

impl Id {
    pub fn new<S: ToString>(id: S, span: Option<Span>) -> Self {
        Self {
            id: id.to_string(),
            span,
        }
    }

    pub fn fmt_err(&self, err_msg: &str) -> String {
        match &self.span {
            Some(span) => span.format(err_msg),
            None => err_msg.to_string(),
        }
    }
}

impl std::fmt::Debug for Id {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        f.debug_struct("Id").field("id", &self.id).finish()
    }
}

/* =================== Impls for Id to make them easier to use ============== */

impl ToString for Id {
    fn to_string(&self) -> String {
        self.id.clone()
    }
}

impl AsRef<str> for Id {
    fn as_ref(&self) -> &str {
        &self.id
    }
}

impl From<&str> for Id {
    fn from(s: &str) -> Self {
        Id {
            id: s.to_string(),
            span: None,
        }
    }
}

impl From<String> for Id {
    fn from(s: String) -> Self {
        Id { id: s, span: None }
    }
}

impl PartialEq<str> for Id {
    fn eq(&self, other: &str) -> bool {
        self.id == other
    }
}

/// Top level AST statement. This contains a list of Component definitions.
#[derive(Clone, Debug)]
pub struct NamespaceDef {
    /// The path to libraries
    pub libraries: Vec<String>,
    /// List of component definitions.
    pub components: Vec<ComponentDef>,
}

/// AST statement for defining components.
#[derive(PartialEq, Eq, Clone, Debug, Derivative)]
#[derivative(PartialOrd, Ord)]
pub struct ComponentDef {
    /// Name of the component.
    pub name: Id,

    /// Defines input and output ports.
    #[derivative(PartialOrd = "ignore")]
    #[derivative(Ord = "ignore")]
    pub signature: Signature,

    /// List of instantiated sub-components
    #[derivative(PartialOrd = "ignore")]
    #[derivative(Ord = "ignore")]
    pub cells: Vec<Cell>,

    /// List of wires
    #[derivative(PartialOrd = "ignore")]
    #[derivative(Ord = "ignore")]
    pub connections: Vec<Connection>,

    /// Single control statement for this component.
    #[derivative(PartialOrd = "ignore")]
    #[derivative(Ord = "ignore")]
    pub control: Control,
}

impl ComponentDef {
    /// Given a Library Context, resolve all the primitive components
    /// in `self` and return the signatures in a HashMap
    pub fn resolve_primitives(
        &self,
        libctx: &LibraryContext,
    ) -> Result<HashMap<Id, Signature>> {
        let mut map = HashMap::new();

        for stmt in &self.cells {
            if let Cell::Prim { data } = stmt {
                let sig = libctx
                    .resolve(&data.instance.name, &data.instance.params)?;
                map.insert(data.name.clone(), sig);
            }
        }

        Ok(map)
    }
}

/// The signature for a component. Contains a list
/// of input ports and a list of output ports.
#[derive(Clone, Debug, Hash, Default, PartialEq, Eq)]
pub struct Signature {
    /// List of input ports.
    pub inputs: Vec<Portdef>,

    /// List of output ports.
    pub outputs: Vec<Portdef>,
}

impl Signature {
    pub fn has_input(&self, name: &str) -> bool {
        self.inputs.iter().any(|e| &e.name == name)
    }

    pub fn has_output(&self, name: &str) -> bool {
        self.outputs.iter().any(|e| &e.name == name)
    }

    pub fn add_input(&mut self, name: &str, width: u64) {
        if self.has_input(name) {
            panic!("signature already has input port: {}", name)
        }
        self.inputs.push((name, width).into());
    }

    pub fn add_output(&mut self, name: &str, width: u64) {
        if self.has_output(name) {
            panic!("signature already has output port: {}", name)
        }
        self.outputs.push((name, width).into());
    }
}

/// The definition of an input/output port.
#[derive(Clone, Debug, Hash, PartialEq, Eq)]
pub struct Portdef {
    /// The name of the port.
    pub name: Id,

    /// The width of the port.
    pub width: u64,
}

/// Helper to construct portdef from str and u64.
impl From<(&str, u64)> for Portdef {
    fn from((name, width): (&str, u64)) -> Self {
        Portdef {
            name: name.into(),
            width,
        }
    }
}

/// Statement that refers to a port on a subcomponent.
/// This is distinct from a `Portdef` which defines a port.
#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
pub enum Port {
    /// Refers to the port named `port` on the subcomponent
    /// `component`.
    Comp { component: Id, port: Id },

    /// Refers to the port named `port` on the component
    /// currently being defined.
    This { port: Id },

    /// `group[name]` parses into `Hole { group, name }`
    /// and is a hole named `name` on group `group`
    Hole { group: Id, name: Id },
}

impl Port {
    /// Returns the name of the port being referenced.
    ///  - `(@ comp A)` returns `A`
    ///  - `(@ this B)` returns `B`
    pub fn port_name(&self) -> &Id {
        match self {
            Port::Comp { port, .. } => port,
            Port::This { port } => port,
            Port::Hole { name, .. } => name,
        }
    }

    /// Returns the edge corresponding to this port in the StructureGraph.
    pub fn get_edge(&self, st: &StructureGraph) -> Result<(NodeIndex, Id)> {
        match self {
            Port::Comp { component, port } => {
                Ok((st.get_node_by_name(component)?, port.clone()))
            }
            Port::This { port } => {
                Ok((st.get_node_by_name(&"this".into()).unwrap(), port.clone()))
            }
            Port::Hole { .. } => Err(Error::MalformedControl(
                "Can't use a hole as a condition.".to_string(),
            )),
        }
    }
}

/// Instantiates a subcomponent named `name` with
/// paramters `params`.
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub struct Compinst {
    /// Name of the subcomponent to instantiate.
    pub name: Id,

    /// List of parameters.
    pub params: Vec<u64>,
}

// ===================================
// AST for wire guard expressions
// ===================================

#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
pub enum NumType {
    Decimal,
    Binary,
    Octal,
    Hex,
}

/// Custom bitwidth numbers
#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
pub struct BitNum {
    pub width: u64,
    pub num_type: NumType,
    pub val: u64,
    pub span: Option<Span>,
}

/// Atomic operations used in guard conditions and RHS of the
/// guarded assignments.
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub enum Atom {
    /// Accessing a particular port on a component.
    Port(Port),
    /// A constant.
    Num(BitNum),
}

/// The AST for GuardExprs
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub enum GuardExpr {
    And(Vec<GuardExpr>),
    Or(Vec<GuardExpr>),
    Eq(Box<GuardExpr>, Box<GuardExpr>),
    Neq(Box<GuardExpr>, Box<GuardExpr>),
    Gt(Box<GuardExpr>, Box<GuardExpr>),
    Lt(Box<GuardExpr>, Box<GuardExpr>),
    Geq(Box<GuardExpr>, Box<GuardExpr>),
    Leq(Box<GuardExpr>, Box<GuardExpr>),
    Not(Box<GuardExpr>),
    Atom(Atom),
}

impl GuardExpr {
    /// Returns all the atoms at the leaves of the guard expression.
    pub fn all_atoms(&self) -> Vec<&Atom> {
        match self {
            GuardExpr::Atom(a) => vec![a],
            GuardExpr::Or(gs) | GuardExpr::And(gs) => {
                gs.iter().map(|g| g.all_atoms()).flatten().collect()
            }
            GuardExpr::Eq(l, r)
            | GuardExpr::Neq(l, r)
            | GuardExpr::Gt(l, r)
            | GuardExpr::Lt(l, r)
            | GuardExpr::Leq(l, r)
            | GuardExpr::Geq(l, r) => {
                let mut atoms = l.all_atoms();
                atoms.append(&mut r.all_atoms());
                atoms
            }
            GuardExpr::Not(g) => g.all_atoms(),
        }
    }

    /// Returns true when this guard is equivalent to the empty guard (true).
    /// Since guards can be arbitrarily complex, this is conservative.
    pub fn provably_true(&self) -> bool {
        match self {
            GuardExpr::Or(es) => es.is_empty(),
            GuardExpr::And(es) => es.is_empty(),
            _ => false,
        }
    }

    /// A convienent constructor for `GuardExpr::And`
    /// that allows chaining construction `g.and(guard)`
    pub fn and_vec(mut atoms: Vec<GuardExpr>) -> Self {
        // Early return if this is a trivial vector.
        if atoms.len() == 1 {
            return atoms.remove(0);
        }

        // Flatten any nested `And` inside the atoms.
        let mut flat_atoms: Vec<GuardExpr> = Vec::with_capacity(atoms.len());
        for atom in atoms {
            match atom {
                GuardExpr::And(mut bs) => flat_atoms.append(&mut bs),
                _ => flat_atoms.push(atom),
            }
        }

        // Remove duplicate elements and any 1s.
        let uniqs = flat_atoms
            .into_iter()
            .unique()
            .filter(|atom| match *atom {
                GuardExpr::Atom(Atom::Num(BitNum { val: 1, .. })) => false,
                _ => true,
            })
            .collect();

        GuardExpr::And(uniqs)
    }

    /// A convienent constructor for `GuardExpr::And`
    /// that allows chaining construction `g.and(guard)`
    pub fn and(lhs: GuardExpr, rhs: GuardExpr) -> Self {
        GuardExpr::and_vec(vec![lhs, rhs])
    }

    pub fn or_vec(mut atoms: Vec<GuardExpr>) -> Self {
        // Early return if this is a trivial vector.
        if atoms.len() == 1 {
            return atoms.remove(0);
        }

        // Flatten nested `Or`
        let mut flat_atoms: Vec<GuardExpr> = Vec::with_capacity(atoms.len());
        for atom in atoms {
            match atom {
                GuardExpr::Or(mut bs) => flat_atoms.append(&mut bs),
                _ => flat_atoms.push(atom),
            }
        }

        // Remove duplicates and any 0s.
        let uniqs = flat_atoms
            .into_iter()
            .unique()
            .filter(|atom| match *atom {
                GuardExpr::Atom(Atom::Num(BitNum { val: 0, .. })) => false,
                _ => true,
            })
            .collect();

        GuardExpr::Or(uniqs)
    }

    /// A convienent constructor for `GuardExpr::And`
    /// that allows chaining construction `g.and(guard)`
    pub fn or(lhs: GuardExpr, rhs: GuardExpr) -> Self {
        GuardExpr::or_vec(vec![lhs, rhs])
    }

    pub fn eq(self, other: GuardExpr) -> Self {
        GuardExpr::Eq(Box::new(self), Box::new(other))
    }

    pub fn neq(self, other: GuardExpr) -> Self {
        GuardExpr::Neq(Box::new(self), Box::new(other))
    }

    pub fn le(self, other: GuardExpr) -> Self {
        GuardExpr::Leq(Box::new(self), Box::new(other))
    }

    pub fn lt(self, other: GuardExpr) -> Self {
        GuardExpr::Lt(Box::new(self), Box::new(other))
    }

    pub fn ge(self, other: GuardExpr) -> Self {
        GuardExpr::Geq(Box::new(self), Box::new(other))
    }

    pub fn gt(self, other: GuardExpr) -> Self {
        GuardExpr::Gt(Box::new(self), Box::new(other))
    }

    pub fn op_str(&self) -> String {
        match self {
            GuardExpr::And(_) => "&".to_string(),
            GuardExpr::Or(_) => "|".to_string(),
            GuardExpr::Eq(_, _) => "==".to_string(),
            GuardExpr::Neq(_, _) => "!=".to_string(),
            GuardExpr::Gt(_, _) => ">".to_string(),
            GuardExpr::Lt(_, _) => "<".to_string(),
            GuardExpr::Geq(_, _) => ">=".to_string(),
            GuardExpr::Leq(_, _) => "<=".to_string(),
            GuardExpr::Not(_) => "!".to_string(),
            GuardExpr::Atom(_) => panic!("No operator string for Atom"),
        }
    }
}

impl BitAnd for GuardExpr {
    type Output = Self;

    fn bitand(self, other: Self) -> Self::Output {
        GuardExpr::and(self, other)
    }
}

impl BitOr for GuardExpr {
    type Output = Self;

    fn bitor(self, other: Self) -> Self::Output {
        GuardExpr::or(self, other)
    }
}

impl Not for GuardExpr {
    type Output = Self;

    fn not(self) -> Self {
        match self {
            GuardExpr::Eq(lhs, rhs) => GuardExpr::Neq(lhs, rhs),
            GuardExpr::Neq(lhs, rhs) => GuardExpr::Eq(lhs, rhs),
            GuardExpr::Gt(lhs, rhs) => GuardExpr::Leq(lhs, rhs),
            GuardExpr::Lt(lhs, rhs) => GuardExpr::Geq(lhs, rhs),
            GuardExpr::Geq(lhs, rhs) => GuardExpr::Lt(lhs, rhs),
            GuardExpr::Leq(lhs, rhs) => GuardExpr::Gt(lhs, rhs),
            GuardExpr::Not(expr) => *expr,
            _ => GuardExpr::Not(Box::new(self)),
        }
    }
}

/// A guard is a conditions in `guard_conj` which guard the value
/// represented by `expr`.
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub struct Guard {
    pub guard: Option<GuardExpr>,
    pub expr: Atom,
}

impl ToString for Atom {
    fn to_string(&self) -> String {
        match self {
            Atom::Port(p) => p.port_name().to_string(),
            Atom::Num(n) => n.val.to_string(),
        }
    }
}

impl ToString for GuardExpr {
    fn to_string(&self) -> String {
        match self {
            GuardExpr::And(branches) => format!(
                "and({})",
                branches
                    .iter()
                    .map(|b| b.to_string())
                    .collect::<Vec<String>>()
                    .join(" ")
            ),
            GuardExpr::Or(branches) => format!(
                "or({})",
                branches
                    .iter()
                    .map(|b| b.to_string())
                    .collect::<Vec<String>>()
                    .join(" ")
            ),
            GuardExpr::Eq(a, b) => {
                format!("{}_eq_{}", a.to_string(), b.to_string())
            }
            GuardExpr::Neq(a, b) => {
                format!("{}_neq_{}", a.to_string(), b.to_string())
            }
            GuardExpr::Gt(a, b) => {
                format!("{}_gt_{}", a.to_string(), b.to_string())
            }
            GuardExpr::Lt(a, b) => {
                format!("{}_lt_{}", a.to_string(), b.to_string())
            }
            GuardExpr::Geq(a, b) => {
                format!("{}_geq_{}", a.to_string(), b.to_string())
            }
            GuardExpr::Leq(a, b) => {
                format!("{}_leq_{}", a.to_string(), b.to_string())
            }
            GuardExpr::Not(a) => format!("!{}", a.to_string()),
            GuardExpr::Atom(a) => a.to_string(),
        }
    }
}

impl ToString for Guard {
    fn to_string(&self) -> String {
        self.guard
            .iter()
            .map(GuardExpr::to_string)
            .collect::<Vec<_>>()
            .join("_")
    }
}

// ===================================
// Data definitions for Structure
// ===================================

/// Data for the `new` structure statement.
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub struct Decl {
    /// Name of the variable being defined.
    pub name: Id,

    /// Name of the component being instantiated.
    pub component: Id,
}

/// Data for the `new-std` structure statement.
#[derive(Clone, Debug, Hash, Eq, PartialEq, PartialOrd, Ord)]
pub struct Prim {
    /// Name of the variable being defined.
    pub name: Id,

    /// Data for instantiating the library component.
    pub instance: Compinst,
}

/// The Cell AST nodes.
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub enum Cell {
    /// Node for instantiating user-defined components.
    Decl { data: Decl },
    /// Node for instantiating primitive components.
    Prim { data: Prim },
}

/// Methods for constructing the structure AST nodes.
impl Cell {
    /// Constructs `Structure::Decl` with `name` and `component`
    /// as arguments.
    pub fn decl(name: Id, component: Id) -> Cell {
        Cell::Decl {
            data: Decl { name, component },
        }
    }

    /// Constructs `Structure::Std` with `name` and `instance`
    /// as arguments.
    pub fn prim(var: Id, prim_name: Id, params: Vec<u64>) -> Cell {
        Cell::Prim {
            data: Prim {
                name: var,
                instance: Compinst {
                    name: prim_name,
                    params,
                },
            },
        }
    }
}

#[allow(clippy::large_enum_variant)]
#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum Connection {
    Group(Group),
    Wire(Wire),
}

#[derive(Derivative, Clone, Debug, Eq)]
#[derivative(PartialEq, PartialOrd, Ord)]
pub struct Group {
    pub name: Id,
    pub wires: Vec<Wire>,
    #[derivative(PartialEq = "ignore")]
    #[derivative(PartialOrd = "ignore")]
    #[derivative(Ord = "ignore")]
    pub attributes: HashMap<String, u64>,
}

/// Data for the `->` structure statement.
#[derive(Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
pub struct Wire {
    /// Source of the wire.
    pub src: Guard,

    /// Guarded destinations of the wire.
    pub dest: Port,
}

// ===================================
// Data definitions for Control Ast
// ===================================

/// Data for the `seq` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct Seq {
    /// List of `Control` statements to run in sequence.
    pub stmts: Vec<Control>,
}

/// Data for the `par` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct Par {
    /// List of `Control` statements to run in parallel.
    pub stmts: Vec<Control>,
}

/// Data for the `if` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct If {
    /// Port that connects the conditional check.
    pub port: Port,

    /// Modules that need to be enabled to send signal on `port`.
    pub cond: Id,

    /// Control for the true branch.
    pub tbranch: Box<Control>,

    /// Control for the true branch.
    pub fbranch: Box<Control>,
}

/// Data for the `if` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct While {
    /// Port that connects the conditional check.
    pub port: Port,

    /// Modules that need to be enabled to send signal on `port`.
    pub cond: Id,

    /// Control for the loop body.
    pub body: Box<Control>,
}

/// Data for the `print` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct Print {
    /// Name of the port to print.
    pub var: Port,
}

/// Data for the `enable` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct Enable {
    /// List of components to run.
    pub comp: Id,
}

/// Data for the `empty` control statement.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub struct Empty {}

/// Control AST nodes.
#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub enum Control {
    /// Represents sequential composition of control statements.
    Seq { data: Seq },
    /// Represents parallel composition of control statements.
    Par { data: Par },
    /// Standard imperative if statement
    If { data: If },
    /// Standard imperative while statement
    While { data: While },
    /// Statement that prints out the value of a port during simulation.
    Print { data: Print },
    /// Runs the control for a list of subcomponents.
    Enable { data: Enable },
    /// Control statement that does nothing.
    Empty { data: Empty },
}

/// Methods for constructing control AST nodes.
#[allow(unused)]
impl Control {
    pub fn seq(stmts: Vec<Control>) -> Control {
        Control::Seq {
            data: Seq { stmts },
        }
    }

    pub fn par(stmts: Vec<Control>) -> Control {
        Control::Par {
            data: Par { stmts },
        }
    }

    pub fn c_if(
        port: Port,
        cond: Id,
        tbranch: Control,
        fbranch: Control,
    ) -> Control {
        Control::If {
            data: If {
                port,
                cond,
                tbranch: Box::new(tbranch),
                fbranch: Box::new(fbranch),
            },
        }
    }

    pub fn c_while(port: Port, cond: Id, body: Control) -> Control {
        Control::While {
            data: While {
                port,
                cond,
                body: Box::new(body),
            },
        }
    }

    pub fn print(var: Port) -> Control {
        Control::Print {
            data: Print { var },
        }
    }

    pub fn enable(comp: Id) -> Control {
        Control::Enable {
            data: Enable { comp },
        }
    }

    pub fn empty() -> Control {
        Control::Empty { data: Empty {} }
    }
}
